

from typing import List
from pdcopyist.src.core.MethodCall import AndOperatorMethodCall, EqualOperatorMethodCall, GetItemMethodCall, MethodCall, OrOperatorMethodCall


class AbcStatement(object):

    def has_exception(self) -> bool:
        raise NotImplementedError

    def to_code(self):
        """
        docstring
        """
        raise NotImplementedError

    def run(self, var_dict):
        """
        docstring
        """
        raise NotImplementedError


class CodeStatement(AbcStatement):

    def __init__(self, code) -> None:
        super().__init__()
        self.code = code
        self.ex = None

    def has_exception(self) -> bool:
        return self.ex is not None

    def to_code(self):
        """
        docstring
        """
        return self.code

    def run(self, var_dict):
        obj = None

        try:
            obj = exec(self.code, {}, var_dict)
        except Exception as ex:
            self.ex = ex
            raise

        return None


class Statement(AbcStatement):
    def __init__(self,  caller_name: str, ret_name: str = None) -> None:
        super().__init__()
        self.calls: List[MethodCall] = []
        self.caller_name = caller_name
        self.ret_name = ret_name

    def __getattr__(self, name):
        call = MethodCall(name)
        self.addCall(call)
        return self

    def __call__(self, *args, **kwargs):
        self.calls[-1].set_args(*args, **kwargs)
        return self

    def __getitem__(self, key):
        call = GetItemMethodCall()
        self.addCall(call)
        call.set_args(key)
        return self

    def __or__(self, other):
        call = OrOperatorMethodCall()
        self.addCall(call)
        call.set_args(other)
        return self

    def __and__(self, other):
        call = AndOperatorMethodCall()
        self.addCall(call)
        call.set_args(other)
        return self

    def __eq__(self, other):
        call = EqualOperatorMethodCall()
        self.addCall(call)
        call.set_args(other)
        return self

    def has_exception(self) -> bool:
        ret = any(c.has_ex for c in self.calls)
        return ret

    def addCall(self, call: MethodCall):
        self.calls.append(call)

    def is_empty(self) -> bool:
        return len(self.calls) == 0

    def get_calls(self):
        return (c for c in self.calls if c.name != '__iter__')

    def to_code(self) -> str:
        codes = (
            c.to_code()
            for c in self.calls
        )

        code = ''.join(codes)

        if self.ret_name:
            return f'{self.ret_name} = {self.caller_name}{code}'

        return f'{self.caller_name}{code}'

    def run(self, var_dict):

        obj = var_dict[self.caller_name]

        for c in self.get_calls():
            try:
                obj = c.run(obj, var_dict=var_dict)
                if self.ret_name:
                    var_dict[self.ret_name] = obj
            except Exception as ex:
                c.set_exception(ex)
                raise



from enum import Enum
from typing import Dict, List
from pdcopyist.src.helper.utils import CanJson
import pandas as pd

_m_ui_col_type_mapping = {
    'object': 'str',
    'int': 'number',
    'int64': 'number',
    'int32': 'number',
    'datetime64[ns]': 'datetime'
}


def to_ui_col_type(org_type):
    stype = str(org_type)
    ret = _m_ui_col_type_mapping.get(stype, stype)
    return ret


class CmdType(Enum):
    Success = 'success'
    Error = 'error'


class FilterItem(CanJson):
    def __init__(self, label, value) -> None:
        self.text = label
        self.value = value


class Field(CanJson):
    @staticmethod
    def from_df(org_df: pd.DataFrame, out_df: pd.DataFrame) -> List['Field']:
        def _get_filters(col: pd.Series) -> List[FilterItem]:

            col_values = col.drop_duplicates().fillna('空格')[:50]

            return [
                FilterItem(v, v) for v in col_values
            ]

        if isinstance(out_df.columns, pd.MultiIndex):
            ret = []
            for k, g in out_df.columns.to_frame().groupby(0, sort=False):
                fat_f = Field(k).set_sorter(False)

                sub_fs = [
                    Field(c, str((k, c))).set_filter(
                        _get_filters(org_df[(k, c)]))
                    for c in g[1]
                ]
                fat_f.set_children(sub_fs)
                ret.append(fat_f)
            return ret

        return [
            Field(c, c).set_filter(_get_filters(org_df[c])) for c in out_df.columns
        ]

    def __init__(self, title: str, field: str = None) -> None:
        self.title: str = title
        self.dataIndex: str = field
        self.children: List[Field] = None
        self.filters: List[FilterItem] = None
        self.showOverflow: bool = True
        self.sorter = True
        self.filterMultiple = True

    def set_children(self, fields: List['Field']):
        self.children = fields

    def set_sorter(self, value: bool) -> 'Field':
        self.sorter = value
        return self

    def set_filter(self, filters: List[FilterItem]) -> 'Field':
        self.filters = filters
        return self


class Table(CanJson):
    @staticmethod
    def from_df(org_df: pd.DataFrame, out_df: pd.DataFrame) -> 'Table':

        fields = Field.from_df(org_df, out_df)

        # df_copy = df.copy()

        if isinstance(out_df.columns, pd.MultiIndex):
            out_df.columns = out_df.columns.to_series().apply(
                lambda x: str(tuple(x)))

        out_df.columns = [str(c) for c in out_df.columns]
        data = out_df.fillna('').to_dict(orient='index')
        data = [
            dict(**{'___key___': k}, **v)
            for k, v in data.items()]

        return Table(fields, data)

    def __init__(self, fields: List[Field], data=[]) -> None:
        self.fields: List[Field] = fields
        self.data: List[Dict] = data

    def addrow(self, data: Dict) -> 'Table':
        self.data.append(data)
        return self

    def set_all_data(self, data: List[Dict]) -> 'Table':
        self.data = data
        return self


class ColumnInfo(CanJson):
    @staticmethod
    def from_df(df: pd.DataFrame) -> List['ColumnInfo']:
        return [
            ColumnInfo(str(c), c, str(t), to_ui_col_type(t))
            for c, t in zip(df.columns, df.dtypes)
        ]

    def __init__(self, field, label, pd_type, ui_type) -> None:
        self.field = field
        self.label = label
        self.type = pd_type
        self.ui_type = ui_type


class PandasCmd(CanJson):
    @staticmethod
    def from_df(org_df: pd.DataFrame, out_df: pd.DataFrame) -> 'PandasCmd':
        table = Table.from_df(org_df, out_df)
        colInfos = ColumnInfo.from_df(org_df)
        return PandasCmd(CmdType.Success, table, colInfos)

    def __init__(self,
                 type: CmdType,
                 table: Table = None,
                 columnInfos: List[ColumnInfo] = []) -> None:
        self.type: CmdType = type
        self.columnInfos: List[ColumnInfo] = columnInfos
        self.table: Table = table or {}



from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args, dt_source_code
from pdcopyist.src.core import UIArgs as ty
import pandas as pd

m_encoding = ['utf8', 'gb2312', 'gbk']


def generate_code(*args, **kwargs):
    encoding = m_encoding[kwargs['encoding']]
    return f'''df.to_csv(r'{kwargs['path']}',encoding='{encoding}',index={kwargs['index']})'''


@dt_source_code(generate_code)
@dt_handle_func('导出文件-csv')
@dt_args(path=ty.Input('文件全路径：'), encoding=ty.Select('编码：', m_encoding, default='utf8'), index=ty.Switch('是否输出行索引：', default=False))
def remove_cols(df: pd.DataFrame, path, encoding, index):
    return df.to_csv(path, encoding=encoding, index=index)





from pdcopyist.src.core.MethodCall import ArgActor
from pdcopyist.src.core import Proxy
from werkzeug.datastructures import FileStorage
from pdcopyist.src.core import UIArgs
from pdcopyist.src.core.UIArgs import AbcArgs, Content
from pdcopyist.src.helper.utils import CanJson
from typing import List


class DataArgs(CanJson):

    def __init__(self, model_type) -> None:
        self.model_type = model_type.__name__
        self.contents: List[Content] = []

    def append_arg(self, args: AbcArgs):
        self.contents.append(args.to_content())


class UIJsonModel(object):

    @staticmethod
    def get_model(file: FileStorage, file_ext: str) -> 'UIJsonModel':
        if file_ext in ['xlsx', 'xls', 'xlsm']:
            return ExcelUIJsonModel(file, file_ext)

        if file_ext in ['csv']:
            return CsvUIJsonModel(file, file_ext)

        if file_ext in ['feather']:
            return FeatherUIJsonModel(file, file_ext)

        raise Exception('not support file')

    def __init__(self, file: FileStorage, file_ext: str) -> None:
        self.file = file
        self.file_ext = file_ext

    def handler(self, caller: Proxy.CallerProxy, file: FileStorage, file_ext: str, **kwargs):
        raise NotImplementedError

    def to_args(self) -> DataArgs:
        raise NotImplementedError


class ExcelUIJsonModel(UIJsonModel):

    s_engine_map = {
        'xls': 'xlrd',
        'xlsx': 'openpyxl',
        'xlsm': 'openpyxl'
    }

    def __init__(self, file, file_ext: str) -> None:
        super().__init__(file, file_ext)

    def _get_sheet_names(self, file):
        import pandas as pd
        xl = pd.ExcelFile(file)
        return xl.sheet_names

    def to_args(self) -> DataArgs:
        pass
        da = DataArgs(ExcelUIJsonModel)
        da.append_arg(UIArgs.Input(
            '文件名：', enable=False, default=self.file.filename))
        da.append_arg(UIArgs.Input('后缀：', enable=False, default=self.file_ext))
        da.append_arg(UIArgs.Select(
            '选择工作表：', self._get_sheet_names(self.file), var_name='sheet_name'))

        return da

    def handler(self, caller: Proxy.CallerProxy, file: FileStorage, file_ext: str, **kwargs):
        if file_ext in self.s_engine_map:
            kwargs['engine'] = self.s_engine_map[file_ext]

        with caller.with_cmd('加载数据') as cmd:
            cmd.create_statement('pd', 'df')
            getattr(caller, 'read_excel')(
                ArgActor(file, str(file.filename)), **kwargs)


class CsvUIJsonModel(UIJsonModel):

    def __init__(self, file, file_ext: str) -> None:
        super().__init__(file, file_ext)

    def to_args(self) -> DataArgs:
        pass
        da = DataArgs(CsvUIJsonModel)
        da.append_arg(UIArgs.Input(
            '文件名：', enable=False, default=str(self.file)))
        da.append_arg(UIArgs.Input('后缀：', enable=False, default=self.file_ext))
        da.append_arg(UIArgs.Select(
            '编码：', source=['utf8', 'gb2312'], var_name='encoding'))

        return da

    def handler(self, caller: Proxy.CallerProxy, file: FileStorage, file_ext: str, **kwargs):
        with caller.with_cmd('加载数据') as cmd:
            cmd.create_statement('pd', 'df')
            getattr(caller, 'read_csv')(
                ArgActor(file, str(file.filename)), **kwargs)


class FeatherUIJsonModel(UIJsonModel):

    def __init__(self, file, file_ext: str) -> None:
        super().__init__(file, file_ext)

    def to_args(self) -> DataArgs:
        pass
        da = DataArgs(FeatherUIJsonModel)
        da.append_arg(UIArgs.Input(
            '文件名：', enable=False, default=str(self.file)))
        da.append_arg(UIArgs.Input('后缀：', enable=False, default=self.file_ext))

        return da

    def handler(self, caller: Proxy.CallerProxy, file: FileStorage, file_ext: str, **kwargs):
        with caller.with_cmd('加载数据') as cmd:
            cmd.create_statement('pd', 'df')
            getattr(caller, 'read_feather')(
                ArgActor(file, str(file.filename)), **kwargs)




import json
from pdcopyist.src.core.UIArgs import Content, AbcArgs
from typing import List

from werkzeug.datastructures import FileStorage
from pdcopyist.src.controllers.loaddata_model import UIJsonModel
import flask
from flask import request, Blueprint
import pandas as pd
from pdcopyist.src.core.Proxy import ProxyManager
from pdcopyist.src.controllers import common

app = Blueprint('loaddata', __name__)


@app.route('/ui_args', methods=['post'])
def get_ui_args():
    '''
    根据文件后缀名，返回界面所需内容
    '''
    file: FileStorage = flask.request.files.get('file')
    ext = flask.request.form['ext']

    model = UIJsonModel.get_model(file, ext)
    ret = model.to_args()
    ret = common.to_json(ret)

    return ret


@app.route('/file', methods=['post'])
def upload_file():
    file: FileStorage = flask.request.files.get('file')

    ext = flask.request.form['ext']
    filename = flask.request.form['file_name']
    args = json.loads(flask.request.form['args'])
    kwargs = {
        a['content']['var_name']: AbcArgs.extract_value(a)
        for a in args['contents']
        if a['content']['var_name'] is not None}

    model = UIJsonModel.get_model(file, ext)

    px = ProxyManager.get()
    px.init_all_proxy()
    px.import_module()

    model.handler(px.pd_read_proxy, file, ext, **kwargs)

    px.read_data()
    return common.df2json(px.get_df_data())

    # todo:根据不同情况做处理
    # px = ProxyManager.get()
    # px.read_data(file, filename, ext, args)
    # return common.df2json(px.get_df_data())



import importlib
from pdcopyist.src.cusFuns.core.DecoratorFuns import dt_handle_func, dt_args, dt_source_code
from pdcopyist.src.core import UIArgs as ty
import pandas as pd


def generate_code(*args, **kwargs):
    kws = {kwargs['org_name']: kwargs['new_name']}
    return f'''df = df.rename(columns={kws})'''


@dt_source_code(generate_code)
@dt_handle_func('修改列名')
@dt_args(org_name=ty.ColumnSelect('待修改列：'), new_name=ty.Input('新列名：', placeholder='输入新的列名'))
def change_column_name(df, org_name, new_name):
    return df.rename(columns={org_name: new_name})





from pdcopyist.src.helper.utils import CanJson
from pdcopyist.src.core import Proxy
from typing import Callable, Dict, Iterable
import sys


class Content(CanJson):
    def __init__(self, title: str, type: str, var_name: str, type_id: str, enable=True, source=None, defaultValue=None) -> None:
        self.title = title
        self.type = type
        self.type_id = type_id
        self.var_name = var_name
        self.enable = enable
        self.source = source
        self.defaultValue = defaultValue


class AbcArgs(object):
    """
    docstring
    """
    pass

    m_type_id = 0
    m_id_type_mapping = {}

    @staticmethod
    def setup_id(class_name: str):
        AbcArgs.m_type_id += 1
        id = str(AbcArgs.m_type_id)
        AbcArgs.m_id_type_mapping[id] = class_name
        return id

    @staticmethod
    def extract_value(content_dict: Dict):
        type_id = content_dict['content']['type_id']
        class_name = AbcArgs.m_id_type_mapping[type_id]
        ct = getattr(sys.modules[__name__], class_name)
        return ct.extract_value(content_dict['content'], content_dict['input'])

    def __init__(self, title, var_name=None, enable=True, default=None, required=True, to_content_overwrite: Callable = None) -> None:
        self.title = title
        self.default = default
        self.required = required
        self.var_name: str = var_name
        self.enable = enable
        self.to_content_overwrite = to_content_overwrite

    def set_var_name(self, var_name) -> 'AbcArgs':
        self.var_name = var_name
        return self

    def to_content(self) -> Content:
        raise NotImplementedError

    def to_rule(self):
        raise NotImplementedError


class ColumnSelect(AbcArgs):

    m_type_id = AbcArgs.setup_id('ColumnSelect')

    @staticmethod
    def extract_value(content_dict: Dict, input):
        """
        docstring
        """
        pass

    def __init__(self, title, var_name=None, enable=True, default=None, required=True, to_content_overwrite: Callable = None) -> None:
        super().__init__(title, var_name=var_name, enable=enable, default=default,
                         required=required, to_content_overwrite=to_content_overwrite)

    def to_content(self) -> Content:
        ct = Content(self.title, 'select', self.var_name, self.m_type_id, enable=self.enable,
                     source='data_columns', defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '必需选择一列', 'trigger': 'change'}]


class Input(AbcArgs):

    m_type_id = AbcArgs.setup_id('Input')

    @staticmethod
    def extract_value(content_dict: Dict, input):
        return input

    def __init__(self, title, var_name=None, placeholder=None, enable=True, default=None, required=True, to_content_overwrite: Callable = None) -> None:
        super().__init__(title, var_name=var_name, enable=enable, default=default,
                         required=required, to_content_overwrite=to_content_overwrite)
        self.placeholder = placeholder

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'input', self.var_name, self.m_type_id, enable=self.enable,
                     defaultValue=self.default)

        ct.placeholder = self.placeholder
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请输入内容', 'trigger': 'blur'}]


class Select(AbcArgs):

    m_type_id = AbcArgs.setup_id('Select')

    @staticmethod
    def extract_value(content_dict: Dict, input):
        source = content_dict['source']
        return source[input]['text']

    def __init__(self, title, source, var_name=None, default=None, required=True) -> None:
        super().__init__(title, var_name=var_name, default=default, required=required)

        if isinstance(source, Iterable):
            source = [{'value': i, 'text': str(v)}
                      for i, v in enumerate(source)]
        self.source = source

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'select', self.var_name, self.m_type_id, enable=self.enable,
                     source=self.source, defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '必需选择一项', 'trigger': 'change'}]


class ColumnWithTypeSelect(AbcArgs):

    m_type_id = AbcArgs.setup_id('ColumnWithTypeSelect')

    @staticmethod
    def extract_value(content_dict: Dict, input):
        pass

    def __init__(self, title, var_name=None, default=None, required=True) -> None:
        super().__init__(title, var_name=var_name, default=default, required=required)

    def get_value(self, index: int):
        df = Proxy.ProxyManager.get().get_last_df_cache()
        col = str(df.columns[index])
        return col

    def to_content(self) -> Content:
        df = Proxy.ProxyManager.get().get_last_df_cache()
        cols = [f'{name}({t})' for name, t in zip(df.columns, df.dtypes)]

        source = [{'value': i, 'text': str(v)}
                  for i, v in enumerate(cols)]

        ct = Content(self.title, 'select', self.var_name, self.m_type_id, enable=self.enable,
                     source=source, defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请输入内容', 'trigger': 'blur'}]


class Switch(AbcArgs):

    m_type_id = AbcArgs.setup_id('Switch')

    @staticmethod
    def extract_value(content_dict: Dict, input):
        return input

    def __init__(self, title, var_name=None, default=None, required=True) -> None:
        super().__init__(title, var_name=var_name, default=default, required=required)

    def to_content(self) -> Content:
        if self.to_content_overwrite:
            return self.to_content_overwrite()

        ct = Content(self.title, 'switch', self.var_name, self.m_type_id, enable=self.enable,
                     defaultValue=self.default)
        return ct

    def to_rule(self):
        return [{'required': self.required, 'message': '请选中一项', 'trigger': 'change'}]
